from fnal_column_analysis_tools.analysis_objects.JaggedCandidateArray import JaggedCandidateArray
from fnal_column_analysis_tools.striped.StripedColumnTransformer import PhysicalColumnGroup
import awkward

def jaggedFromColumnGroup(cgroup):
    if isinstance(cgroup,PhysicalColumnGroup):
        return JaggedCandidateArray.candidatesfromcounts(counts = cgroup.counts(),
                                                         p4 = cgroup.p4Column(),
                                                         **cgroup.otherColumns())
    else:
        return awkward.JaggedArray.fromcounts(cgroup.counts(),
                                              awkward.Table(cgroup.columns()))
                                                    



import numpy as np
from awkward.array.jagged import JaggedArray
from copy import deepcopy

class denselookup(object):
    def __init__(self,values,dims): 
        self.__dimension = 0
        whattype = type(dims)
        if whattype == np.ndarray:
            self.__dimension = 1
        else:
            self.__dimension = len(dims)        
        if self.__dimension == 0:
            raise Exception('Could not define dimension for {}'.format(whattype))
        self.__axes = deepcopy(dims)
        self.__values = deepcopy(values)
        self.__type = type(self.__values)
    
    def __call__(self,*args):        
        inputs = list(args)
        offsets = None
        # TODO: check can use offsets (this should always be true for striped)
        # Alternatively we can just use starts and stops
        for i in range(len(inputs)):
            if isinstance(inputs[i], JaggedArray):
                if offsets is not None and offsets.base is not inputs[i].offsets.base:
                    if type(offsets) is int:
                        raise Exception('do not mix JaggedArrays and numpy arrays when calling denselookup')
                    else:
                        raise Exception('All input jagged arrays must have a common structure (offsets)!')
                offsets = inputs[i].offsets
                inputs[i] = inputs[i].content
            elif isinstance(inputs[i], np.ndarray):
                offsets = -1
        retval = self.__evaluate(*tuple(inputs))
        if offsets is not None and type(offsets) is not int:
            retval = JaggedArray.fromoffsets(offsets,retval)
        return retval
                                               
    
    def __evaluate(self,*args):        
        indices = [] 
        for arg in args: 
            if type(arg) == JaggedArray: raise Exception('JaggedArray in inputs')
        if self.__dimension == 1:
            indices.append(np.clip(np.searchsorted(self.__axes, args[0], side='right')-1,0,self.__values.shape[0]-1))
        else:
            for dim in range(self.__dimension):
                #print self.__axes[dim], self.__values.shape
                indices.append(np.clip(np.searchsorted(self.__axes[dim], args[dim], side='right')-1,0,self.__values.shape[len(self.__axes)-dim-1]-1))
        indices.reverse()
        return self.__values[tuple(indices)]
    
    def __repr__(self):
        myrepr = "{} dimensional histogram with axes:\n".format(self.__dimension)
        temp = "" 
        if self.__dimension == 1:
            temp = "\t1: {}\n".format(self.__axes)
        else:
            temp = "\t1: {}\n".format(self.__axes[0])
        for idim in range(1,self.__dimension):
            temp += "\t{}: {}\n".format(idim+1,self.__axes[idim])        
        myrepr += temp
        return myrepr

class evaluator(object):
    def __init__(self,names,primitives):
        self.__functions = {}
        for key in names.keys():
            self.__functions[key] = denselookup(*primitives[names[key]])
            
    def __dir__(self):
        return self.__functions.keys()
        
    def __getitem__(self, key):
        return self.__functions[key]




# here we want to define a group of columns from striped
# that all pertain to the same object type
# i.e. group = PhysicalColumnGroup(job,p4="Electron_p4",""NElectrons","Electron_p4","Electron_id1",...)
# this is specialized to striped so far
# we should try to peel it apart for other things

class ColumnGroup(object):
    def __init__(self,events,objName,*args):
        self.__map = {}        
        eventObj = getattr(events,objName)
        self.__counts = getattr(events,objName).count        
        for arg in args:
            callStack = arg.split('.')
            retval = getattr(eventObj,callStack[0])
            for i in range(1,len(callStack)):
                retval = getattr(retval,callStack[i])
            self.__map[arg] = retval
            
    def __getitem__(self,name):
        return self.__map[name]
    
    def columnsWithout(self,toremove):
        out = {}
        out.update(self.__map)
        if isinstance(toremove,str):
            del out[toremove]
        else:
            for key in toremove:
                del out[key]
        return out
    
    def columns(self):
        return self.__map
    
    def counts(self):
        return self.__counts
    
class PhysicalColumnGroup(ColumnGroup):
    def __init__(self,events,objName,p4Name,*args):
        self.__p4  = p4Name
        allargs = [p4Name]
        allargs.extend(args)        
        super(PhysicalColumnGroup,self).__init__(events,objName,*allargs)
        if p4Name is not None:
            self.setP4Name(p4Name)
    
    def setP4Name(self,name):
        if name not in self.columns().keys():
            raise Exception('{} not an available name in this PhysicalColumnGroup'.format(name))
        self.__p4 = name
    
    def p4Name(self):
        if self.__p4 is None:
            raise Exception('p4 is not set for this PhysicalColumnGroup')
        return self.__p4
    
    def p4Column(self):        
        return self[self.p4Name()]
    
    def otherColumns(self):
        return self.columnsWithout(self.p4Name())




import uproot
from fnal_column_analysis_tools.lookup_tools.evaluator import evaluator

TH1D = "<class 'uproot.rootio.TH1D'>"
TH2D = "<class 'uproot.rootio.TH2D'>"
TH1F = "<class 'uproot.rootio.TH1F'>"
TH2F = "<class 'uproot.rootio.TH2F'>"
RootDir = "<class 'uproot.rootio.ROOTDirectory'>"

def convert_root_file(file):
    converted_file = {}
    dumFile = uproot.open(file.strip())
    for i, key in enumerate(dumFile.keys()):
        histType = str(type(dumFile[key[:-2]]))
        if histType == TH1D or histType == TH2D or histType==TH1F or histType==TH2F:
            if  not ("bound method" in str(dumFile[key[:-2]].numpy)):
                converted_file[key[:-2]] = dumFile[key[:-2]].numpy
            else:
                converted_file[key[:-2]] = dumFile[key[:-2]].numpy()
        elif histType == RootDir: #means there are subdirectories wihin main directory
            for j, key2 in enumerate(dumFile[key[:-2]].keys()):
                histType2 = str(type(dumFile[key[:-2]][key2[:-2]]))
                if histType2 == TH1D or histType2 == TH2D or histType2==TH1F or histTyp2==TH2F:
                    if not ("bound method" in str(dumFile[key[:-2]][key2[:-2]].numpy)):
                        converted_file[key[:-2]+'/'+key2[:-2]] = dumFile[key[:-2]][key2[:-2]].numpy
                    else:
                        converted_file[key[:-2]+'/'+key2[:-2]] = dumFile[key[:-2]][key2[:-2]].numpy()           
        else:
            tempArrX= dumFile[key[:-2]]._fEX
            tempArrY= dumFile[key[:-2]]._fEY
            converted_file[key[:-2]] = [tempArrX, tempArrY]
    return converted_file

def csvToVec(csvFilePath):
    p = csvFilePath
    f = open(p).readlines()
    
    retVec = []
    for line in f:
        line = re.sub(',', '', line)
        vec = line.split()
        retVec.append(vec)
    retVec.pop(0)
    for i, e in enumerate(retVec):
        retVec[i][0]= int(e[0])
        retVec[i][3]=int(e[3])
        retVec[i][4]=float(e[4])
        retVec[i][5]=float(e[5])
        retVec[i][6]=float(e[6])
        retVec[i][7]=float(e[7])
        retVec[i][8]=float(e[8])
        retVec[i][9]=float(e[9])
        retVec[i][10]=eval(e[10])
        
    return retVec

def makeSfSpecificVec(csvVec,OP, mType,sysType, flavor):
    retVec = []
    for i,e in enumerate(csvVec):
        if e[0] == OP and e[1] ==mType and e[2]==sysType and e[3] == flavor:
            e.pop(0)
            e.pop(0)
            e.pop(0)
            e.pop(0)
            retVec.append(e)
    
    return retVec

file_converters = {'root':convert_root_file,
                   'csv':csvToVec}

class extractor(object):
    def __init__(self):
        self.__weights = []
        self.__names = {}
        self.__filecache = {}
        self.__finalized = False
    
    def add_weight_set(self,local_name,weights):
        if self.__finalized: 
            raise Exception('extractor is finalized cannot add new weights!')
        if local_name in self.__names.keys():
            raise Exception('weights name "{}" already defined'.format(local_name))
        self.__names[local_name] = len(self.__weights)
        self.__weights.append(weights)
    
    def add_weight_sets(self,weightsdescs):
        # expect file to be formatted <local name> <name> <weights file>
        # allow * * <file> and <prefix> * <file> to do easy imports of whole file
        for weightdesc in weightsdescs:
            temp = weightdesc.split(' ')
            if len(temp) != 3: 
                raise Exception('"{}" not formatted as "<local name> <name> <weights file>"'.format(weightdesc))
            (local_name,name,file) = tuple(temp)
            if name == '*':
                self.import_file(file)
                weights = self.__filecache[file]
                for key, value in weights.iteritems():
                    if local_name == '*':
                        self.add_weight_set(key,value)
                    else:
                        self.add_weight_set(local_name+key,value)
            else:
                weights = self.extract_from_file(file,name)
                self.add_weight_set(local_name,weights)
    
    def import_file(self,file):
        if file not in self.__filecache.keys():
            self.__filecache[file] = file_converters[file.split('.')[-1].strip()](file)
    
    def extract_from_file(self, file, name):        
        self.import_file(file)
        weights = self.__filecache[file]
        bname = name.encode()
        if bname not in weights.keys(): 
            print(weights.keys())
            raise Exception('Weights named "{}" not in {}!'.format(name,file))        
        return weights[bname]
          
    def finalize(self):
        if self.__finalized: 
            raise Exception('extractor is already finalized!')
        del self.__filecache
        self.__finalized = True
    
    def make_evaluator(self):
        return evaluator(self.__names,self.__weights)



#!/usr/bin/env python

#BSD 3-Clause License
#
#Copyright (c) 2018, Fermilab
#All rights reserved.
#
#Redistribution and use in source and binary forms, with or without
#modification, are permitted provided that the following conditions are met:
#
#* Redistributions of source code must retain the above copyright notice, this
#  list of conditions and the following disclaimer.
#
#* Redistributions in binary form must reproduce the above copyright notice,
#  this list of conditions and the following disclaimer in the documentation
#  and/or other materials provided with the distribution.
#
#* Neither the name of the copyright holder nor the names of its
#  contributors may be used to endorse or promote products derived from
#  this software without specific prior written permission.
#
#THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import fnal_column_analysis_tools.lookup_tools
import fnal_column_analysis_tools.analysis_objects
import fnal_column_analysis_tools.striped




import time

class Timer(object):
    def __init__(self):
        self.__tic = {}
        self.__toc = {}

    def set_tic(self,name):
        self.__tic[name] = time.time()

    def set_toc(self,name):
        self.__toc[name] = time.time()

    def get_tic(self,name):
        return self.__tic[name]

    def get_toc(self,name):
        return self.__toc[name]

    def tics(self):
        return self.__tic

    def tocs(self):
        return self.__toc
    
    def fill_job_timer_info(self,stripelen,job):
        tic = self.__tic
        toc = self.__toc
        for key in tic.keys():
            dtime = toc[key]-tic[key]
            job.fill(
                category=key,
                stripeThroughput=stripelen/(dtime+1e-8),
                stripeTime=dtime
            )



from fnal_column_analysis_tools.striped.StripedColumnTransformer import ColumnGroup
from fnal_column_analysis_tools.striped.StripedColumnTransformer import PhysicalColumnGroup
from fnal_column_analysis_tools.striped.ColumnGroup2JaggedTable import jaggedFromColumnGroup



import awkward
import uproot_methods
import numpy as np

JaggedTLorentzVectorArray = awkward.Methods.mixin(uproot_methods.classes.TLorentzVector.ArrayMethods, awkward.JaggedArray)

#functions to quickly cash useful quantities
def fast_pt(p4):
    return np.hypot(p4.x,p4.y)

def fast_eta(p4):
    px = p4.x
    py = p4.y
    pz = p4.z
    p3mag = np.sqrt(px*px + py*py + pz*pz)
    return np.arctanh(pz/p3mag)

def fast_phi(p4):
    return np.arctan2(p4.y,p4.x)

def fast_mass(p4):
    px = p4.x
    py = p4.y
    pz = p4.z
    en = p4.t
    p3mag2 = (px*px + py*py + pz*pz)
    return np.sqrt(np.abs(en*en - p3mag2))

class JaggedCandidateMethods(awkward.Methods):
    
    @classmethod
    def candidatesfromcounts(cls,counts,p4,**kwargs):
        offsets = awkward.array.jagged.counts2offsets(counts)
        return cls.candidatesfromoffsets(offsets,p4,**kwargs)
    
    @classmethod
    def candidatesfromoffsets(cls,offsets,p4,**kwargs):
        items = {}
        if isinstance(p4,uproot_methods.TLorentzVectorArray):
            items['p4'] = p4
        else:
            items['p4'] = uproot_methods.TLorentzVectorArray(p4[:,0],p4[:,1],
                                                             p4[:,2],p4[:,3])
        thep4 = items['p4']
        items['__fast_pt'] = fast_pt(thep4)
        items['__fast_eta'] = fast_eta(thep4)
        items['__fast_phi'] = fast_phi(thep4)
        items['__fast_mass'] = fast_mass(thep4)
        items.update(kwargs)
        return cls.fromoffsets(offsets,awkward.Table(items))
    
    @property
    def p4(self):
        return self['p4']
    
    @property
    def pt(self):
        return self['__fast_pt']
    
    @property
    def eta(self):
        return self['__fast_eta']
    
    @property
    def phi(self):
        return self['__fast_phi']
    
    @property
    def mass(self):
        return self['__fast_mass']
    
    def at(self,what):
        thewhat = super(JaggedCandidateMethods,self).at(what)
        if 'p4' in thewhat.columns:
            return self.fromjagged(thewhat)
        return thewhat
    
    def pairs(self, same=True):
        outs = super(JaggedCandidateMethods, self).pairs(same)
        outs['p4'] = outs.at(0)['p4'] + outs.at(1)['p4']
        thep4 = outs['p4']
        outs['__fast_pt'] = awkward.JaggedArray.fromoffsets(outs.offsets,fast_pt(thep4.content))
        outs['__fast_eta'] = awkward.JaggedArray.fromoffsets(outs.offsets,fast_eta(thep4.content))
        outs['__fast_phi'] = awkward.JaggedArray.fromoffsets(outs.offsets,fast_phi(thep4.content))
        outs['__fast_mass'] = awkward.JaggedArray.fromoffsets(outs.offsets,fast_mass(thep4.content))
        return self.fromjagged(outs)
    
    def cross(self, other):
        outs = super(JaggedCandidateMethods, self).cross(other)
        #currently JaggedArray.cross() has some funny behavior when it encounters the
        # p4 column and makes some wierd new column... for now I just delete it and reorder
        # everything looks ok after that
        keys = outs.columns
        reorder = False
        for key in keys:
            if not isinstance(outs[key].content,awkward.array.table.Table):
                del outs[key]
                reorder = True
        if reorder:
            keys = outs.columns
            realkey = {}
            for i in xrange(len(keys)):
                realkey[keys[i]] = str(i)
            for key in keys:
                if realkey[key] != key:
                    outs[realkey[key]] = outs[key]
                    del outs[key]
        keys = outs.columns
        for key in keys:
            if 'p4' not in outs.columns:
                outs['p4'] = outs[key]['p4']
            else:
                outs['p4'] = outs['p4'] + outs[key]['p4']
        thep4 = outs['p4']
        outs['__fast_pt'] = awkward.JaggedArray.fromoffsets(outs.offsets,fast_pt(thep4.content))
        outs['__fast_eta'] = awkward.JaggedArray.fromoffsets(outs.offsets,fast_eta(thep4.content))
        outs['__fast_phi'] = awkward.JaggedArray.fromoffsets(outs.offsets,fast_phi(thep4.content))
        outs['__fast_mass'] = awkward.JaggedArray.fromoffsets(outs.offsets,fast_mass(thep4.content))
        return self.fromjagged(outs)
        
    def __getattr__(self,what):
        if what in self.columns:
            return self[what]
        return getattr(super(JaggedCandidateMethods,self),what)

JaggedCandidateArray = awkward.Methods.mixin(JaggedCandidateMethods, awkward.JaggedArray)
